import warnings
import pandas as pd


def assert_dict(actual: dict, expected: dict, not_tested_keys: list = []):
    for key, value in actual.items():
        if key not in not_tested_keys:
            if isinstance(value, list):
                assert expected[key] == value
            elif isinstance(value, pd.DataFrame):
                assert_dataframe(actual=value, expected=expected[key])
            elif isinstance(value, dict):
                assert value == expected[key]
            else:
                warnings.warn(
                    "test coverage not implemented for this {} type".format(type(value))
                )


def assert_dataframe(actual: pd.DataFrame, expected: pd.DataFrame):
    actual = actual.reindex(expected.index)
    actual = actual[expected.columns]
    pd.testing.assert_frame_equal(actual, expected, check_dtype=False)



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

# Formating
def set_layout_and_display_2y(
    figure_object,
    x_col_name: str,
    y1_col_name: str,
    y2_col_name: str,
    title_str: str,
    y1_is_percentage: bool,
    y2_is_percentage: bool,
    x_is_int: bool,
    x_is_date: bool,
    size_width: int,
    size_height: int,
    are_negative_numbers_possible: bool = False,
    max_abs_value_main_y: float = None,
    max_abs_value_secondary_y: float = None,
    y_bound_multiple: float = 1.1,
    barmode_type: str = "relative",
):

    # Set y as percentage if required
    if y1_is_percentage:
        figure_object.layout.yaxis.tickformat = ",.0%"
    if y2_is_percentage:
        figure_object.layout.yaxis2.tickformat = ",.0%"

    # Set proper format for x axis
    if x_is_int:
        figure_object.layout.xaxis.tickformat = ",d"
    elif x_is_date:
        figure_object.update_layout(xaxis=dict(tickformat="%m-%Y"))

    # Set legend position and size
    figure_object.update_layout(
        width=size_width,
        height=size_height,
        xaxis_title=x_col_name,
        title=title_str,
        barmode=barmode_type,
    )

    # Set y axis names
    figure_object.update_yaxes(title_text=y1_col_name, secondary_y=False)
    figure_object.update_yaxes(title_text=y2_col_name, secondary_y=True)

    # Align 0 - This will work as long as there are no negative values
    figure_object.update_yaxes(rangemode="tozero")
    figure_object.update_xaxes(rangemode="tozero")

    # Update y axis range if required
    if are_negative_numbers_possible:

        if max_abs_value_secondary_y is None:
            raise ValueError("Pls provide abs max value for secondary y axis")
        if max_abs_value_main_y is None:
            raise ValueError("Pls provide abs max value for primary y axis")

        figure_object.update_yaxes(
            scaleanchor="y1",
            scaleratio=1,
            constraintoward="bottom",
            secondary_y=False,
            range=[
                max_abs_value_main_y * -y_bound_multiple,
                max_abs_value_main_y * y_bound_multiple,
            ],
        )

        figure_object.update_yaxes(
            scaleanchor="y2",
            scaleratio=1,
            constraintoward="bottom",
            secondary_y=True,
            range=[
                max_abs_value_secondary_y * -y_bound_multiple,
                max_abs_value_secondary_y * y_bound_multiple,
            ],
        )

    # https://community.plotly.com/t/plotly-express-with-xaxis-having-integers-strings/34777/2
    figure_object.update_layout(xaxis_type="category")

    return figure_object


def set_layout_and_display_1y(
    figure_object,
    x_col_name,
    y_col_name,
    title_str,
    y_is_percentage,
    showlegend,
    show_y_axis,
    show_x_axis,
    size_width,
    size_height,
    y_axis_range,
):
    """
    Sets the layout and displays the chart
    """

    # Set the layout
    figure_object.update_layout(
        title=title_str,
        xaxis_title=x_col_name,
        yaxis_title=y_col_name,
        showlegend=showlegend,
        yaxis={"visible": show_y_axis},
        xaxis={"visible": show_x_axis},
        width=size_width,
        height=size_height,
        yaxis_range=y_axis_range,
    )

    # Set the y axis to be percentage
    if y_is_percentage:
        figure_object.update_yaxes(tickformat=",.0%")

    return figure_object


## MAIN FUNCTIONS ##


def standard_chart_formatting_1y(func):
    """
    Decorator for adding formatting on top of charts
    """

    def inner(*args, **kwargs):
        fig = func(*args, **kwargs)

        # Lets bypass the decorator if certain conditions are met
        bypass_decorator = False
        for key in ["marginal", "facet_col", "facet_row"]:
            if key in kwargs:
                bypass_decorator = True

        if bypass_decorator:
            return fig

        # x_title_text
        x_title_text = kwargs["x_title_text"] if "x_title_text" in kwargs else None
        # y_title_text
        y_title_text = kwargs["y_title_text"] if "y_title_text" in kwargs else None
        # title
        title = kwargs["title"] if "title" in kwargs else None
        # y_is_percentage
        y_is_percentage = (
            kwargs["y_is_percentage"] if "y_is_percentage" in kwargs else False
        )
        # showlegend
        showlegend = kwargs["showlegend"] if "showlegend" in kwargs else True
        # show_y_axis
        show_y_axis = kwargs["show_y_axis"] if "show_y_axis" in kwargs else True
        # show_x_axis
        show_x_axis = kwargs["show_x_axis"] if "show_x_axis" in kwargs else True
        # size_width
        size_width = kwargs["size_width"] if "size_width" in kwargs else 1000
        # Size height
        size_height = kwargs["size_height"] if "size_height" in kwargs else 500

        fig = set_layout_and_display_1y(
            figure_object=fig,
            x_col_name=x_title_text,
            y_col_name=y_title_text,
            title_str=title,
            y_is_percentage=y_is_percentage,
            showlegend=showlegend,
            show_y_axis=show_y_axis,
            show_x_axis=show_x_axis,
            size_width=size_width,
            size_height=size_height,
            y_axis_range=None,
        )
        return fig

    return inner


def standard_chart_formatting_2y(func):
    """
    Decorator for adding formatting on top of charts
    """

    def inner(*args, **kwargs):
        fig = func(*args, **kwargs)

        # x_title_text
        x_title_text = kwargs["x_title_text"] if "x_title_text" in kwargs else None
        # y_title_text
        y1_title_text = kwargs["y1_title_text"] if "y1_title_text" in kwargs else None
        y2_title_text = kwargs["y2_title_text"] if "y2_title_text" in kwargs else None
        # title
        title = kwargs["title"] if "title" in kwargs else None
        # y_is_percentage
        y1_is_percentage = (
            kwargs["y1_is_percentage"] if "y1_is_percentage" in kwargs else False
        )
        y2_is_percentage = (
            kwargs["y2_is_percentage"] if "y2_is_percentage" in kwargs else False
        )
        # showlegend
        showlegend = kwargs["showlegend"] if "showlegend" in kwargs else True
        # show_y_axis
        show_y1_axis = kwargs["show_y1_axis"] if "show_y1_axis" in kwargs else True
        show_y2_axis = kwargs["show_y2_axis"] if "show_y2_axis" in kwargs else True
        # show_x_axis
        show_x_axis = kwargs["show_x_axis"] if "show_x_axis" in kwargs else True
        # size_width
        size_width = kwargs["size_width"] if "size_width" in kwargs else 1000
        # Size height
        size_height = kwargs["size_height"] if "size_height" in kwargs else 500
        # others
        x_is_int = kwargs["x_is_int"] if "x_is_int" in kwargs else False
        x_is_date = kwargs["x_is_date"] if "x_is_date" in kwargs else not x_is_int
        are_negative_numbers_possible = (
            kwargs["are_negative_numbers_possible"]
            if "are_negative_numbers_possible" in kwargs
            else False
        )
        max_abs_value_main_y = (
            kwargs["max_abs_value_main_y"] if "max_abs_value_main_y" in kwargs else None
        )
        max_abs_value_secondary_y = (
            kwargs["max_abs_value_secondary_y"]
            if "max_abs_value_secondary_y" in kwargs
            else None
        )
        y_bound_multiple = (
            kwargs["y_bound_multiple"] if "y_bound_multiple" in kwargs else 1.1
        )

        fig = set_layout_and_display_2y(
            figure_object=fig,
            x_col_name=x_title_text,
            y1_col_name=y1_title_text,
            y2_col_name=y2_title_text,
            title_str=title,
            y1_is_percentage=y1_is_percentage,
            y2_is_percentage=y2_is_percentage,
            x_is_int=x_is_int,
            x_is_date=x_is_date,
            size_width=size_width,
            size_height=size_height,
            are_negative_numbers_possible=are_negative_numbers_possible,
            max_abs_value_main_y=max_abs_value_main_y,
            max_abs_value_secondary_y=max_abs_value_secondary_y,
            y_bound_multiple=y_bound_multiple,
        )
        return fig

    return inner


# Function to create histogram plot using plotly
@standard_chart_formatting_1y
def histogram_plotly(
    df: pd.DataFrame,
    x_col_name: str,
    y_col_name: str,
    nbins: int,
    color: str = None,
    **kwargs,
) -> None:
    """
    color can be used to plot different categories separately - For example same plot for different sex
    """
    fig = px.histogram(df, x=x_col_name, y=y_col_name, color=color, nbins=nbins)

    return fig


@standard_chart_formatting_1y
def line_plotly(
    df: pd.DataFrame,
    x_col_name: str,
    hover_data: list,
    **kwargs,
) -> None:
    """
    Options for marginal include "histogram", "rug"
    """
    fig = px.line(df, x=x_col_name, y="cdf", hover_data=hover_data)
    return fig


@standard_chart_formatting_1y
def plotly_heatmap(
    _df: pd.DataFrame,
    annotation: bool = True,
    **kwargs,
) -> None:
    df = _df.copy()
    fig = px.imshow(
        df.round(2),
        color_continuous_scale="RdBu",
        origin="lower",
        text_auto=annotation,
    )
    """
    fig.update_layout(showlegend=False)
    fig.update_yaxes(visible=show_y_axis)
    fig.update_layout(
        autosize=False,
        width=size,
        height=size,
    )
    """
    return fig


@standard_chart_formatting_2y
def chart_contribution_effect_with_negative_values_and_2y(
    _df: pd.DataFrame,
    bars_positive_list: list,
    bars_negative_list: list,
    lines_list: list,
    round_y2_num_digits=2,
    **kwargs,
):
    """ """
    df = _df.copy()
    x = df.index.astype(str)

    fig = go.Figure()
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    bar_sum_pos = 0
    for col in bars_positive_list:
        fig.add_trace(
            go.Bar(x=x, y=df[col].values, name=col, showlegend=True, yaxis="y1")
        )
        bar_sum_pos += df[col].fillna(0).values

    bar_sum_neg = 0
    for col in bars_negative_list:
        fig.add_trace(
            go.Bar(x=x, y=-df[col].values, name=col, showlegend=True, yaxis="y1")
        )
        bar_sum_neg += df[col].fillna(0).values

    # Largest magnitude in line plots for use in setting bounds
    max_abs_value_secondary = 0
    for col in lines_list:
        fig.add_trace(
            go.Scatter(
                x=x,
                y=df[col].values,
                name=col,
                showlegend=True,
                yaxis="y2",
                hovertemplate="%{y:." + str(round_y2_num_digits) + "f}",
            )
        )
        max_abs_value_secondary = max(
            max_abs_value_secondary, max(abs(df[col].dropna()))
        )

    return fig


@standard_chart_formatting_2y
def chart_bar_and_lines(
    _df: pd.DataFrame,
    bar_column: str,
    line_columns_list: list,
    are_negative_numbers_possible: bool = False,
    max_abs_value_main_y: float = None,
    max_abs_value_secondary_y: float = None,
    **kwargs,
):
    """ """
    df_output = _df.copy()

    # make index as str so that it is JSON serializable
    df_output.index = df_output.index.astype(str)

    # Create figure with secondary y-axis
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    # Add traces
    fig.add_trace(
        go.Bar(x=df_output.index, y=df_output[bar_column].values, name=bar_column),
        secondary_y=False,
    )

    for line_column in line_columns_list:
        fig.add_trace(
            go.Scatter(
                x=df_output.index, y=df_output[line_column].values, name=line_column
            ),
            secondary_y=True,
        )

    # if numbers could be negative then make some changes
    if are_negative_numbers_possible:
        # find max_abs_value_main_y
        max_abs_value_main_y_actual = df_output[bar_column].abs().max()
        if max_abs_value_main_y is None:
            max_abs_value_main_y = max_abs_value_main_y_actual

        # find max_abs_value_secondary_y
        max_abs_value_secondary_y_actual = (
            df_output[line_columns_list].abs().max(axis=1).max(axis=0)
        )
        if max_abs_value_secondary_y is None:
            max_abs_value_secondary_y = max_abs_value_secondary_y_actual

    return fig


@standard_chart_formatting_1y
def chart_cohort(
    _df: pd.DataFrame,
    cohort_col_name: str,
    x_axis_col_name: str,
    color_col_name: str,
    line_to_be_dashed: str = None,
    **kwargs,
):
    df = _df.copy()
    df = df.dropna(axis=0, how="all")
    df = df.reset_index()
    df = pd.melt(df, id_vars=cohort_col_name).reset_index()

    # create dashed line if required
    if line_to_be_dashed is not None:
        temp_col_name = "dashed_or_not"
        non_dash_line = 0
        dash_line = 3
        df[temp_col_name] = non_dash_line
        df.loc[df[cohort_col_name] == line_to_be_dashed, temp_col_name] = dash_line
    else:
        temp_col_name = ""

    df[x_axis_col_name] = df[x_axis_col_name].astype(str)
    fig = (
        px.line(
            df,
            x=x_axis_col_name,
            y="value",
            color=color_col_name,
            line_dash=temp_col_name,
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(cohort_col_name + "=", ""))
        )
        .for_each_trace(lambda t: t.update(name=t.name.replace("00:00:00", "")))
        .for_each_trace(lambda t: t.update(name=t.name.replace(temp_col_name, "")))
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(", =" + str(non_dash_line), ""))
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(", =" + str(dash_line), ""))
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(color_col_name + "=", ""))
        )
        .for_each_trace(lambda t: t.update(name=t.name.replace(", 0", "")))
        .for_each_trace(lambda t: t.update(name=t.name.replace(", 3", "")))
    )

    return fig


@standard_chart_formatting_2y
def chart_cohort_2y(
    _df: pd.DataFrame,
    _df_ct_ue: pd.DataFrame,
    cohort_col_name: str,
    x_axis_col_name: str,
    x2_axis_col_name: str,
    y2_axis_col_name: str,
    color_col_name: str,
    line_to_be_dashed: str = None,
    **kwargs,
):
    df = _df.copy()
    # df = df.dropna(axis=0, how="all")
    df = df.reset_index()
    df = pd.melt(df, id_vars=cohort_col_name).reset_index()
    df_ct_ue = _df_ct_ue.copy()

    # Sort values
    df = df.sort_values(by=[cohort_col_name], ascending=True)
    df_ct_ue = df_ct_ue.sort_values(by=[x2_axis_col_name], ascending=True)

    # create dashed line if required
    if line_to_be_dashed is not None:
        temp_col_name = "dashed_or_not"
        non_dash_line = 0
        dash_line = 3
        df[temp_col_name] = non_dash_line
        df.loc[df[cohort_col_name] == line_to_be_dashed, temp_col_name] = dash_line

    # primary axis
    df[x_axis_col_name] = df[x_axis_col_name].astype(str)
    fig = (
        px.line(
            df,
            x=x_axis_col_name,
            y="value",
            color=color_col_name,
            line_dash=temp_col_name,
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(cohort_col_name + "=", ""))
        )
        .for_each_trace(lambda t: t.update(name=t.name.replace("00:00:00", "")))
        .for_each_trace(lambda t: t.update(name=t.name.replace(temp_col_name, "")))
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(", =" + str(non_dash_line), ""))
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(", =" + str(dash_line), ""))
        )
        .for_each_trace(
            lambda t: t.update(name=t.name.replace(color_col_name + "=", ""))
        )
    )

    # secondary axis
    df_ct_ue[x2_axis_col_name] = df_ct_ue[x2_axis_col_name].astype(str)
    fig2 = px.bar(df_ct_ue, x=x2_axis_col_name, y=y2_axis_col_name)
    fig2.update_traces(opacity=0.2)
    fig2.update_traces(yaxis="y2")

    # Combine the two together
    subfig = make_subplots(specs=[[{"secondary_y": True}]])
    subfig.add_traces(fig.data + fig2.data)

    return subfig


@standard_chart_formatting_1y
def stacked_bar_chart(
    _df: pd.DataFrame,
    **kwargs,
):
    """ """
    df = _df.copy()
    df.index = df.index.astype(str)

    x = df.index
    fig = go.Figure()
    for col in df.columns:
        fig.add_trace(go.Bar(x=x, y=df[col].values, name=col))

    fig.update_layout(barmode="relative")
    return fig


@standard_chart_formatting_1y
def multiple_line_charts(
    _df: pd.DataFrame,
    x_col_name: str,
    y_col_name: str,
    color_col_name: str = None,
    hover_col_list: list = None,
    **kwargs,
):
    """ """
    df = _df.copy()

    if color_col_name is not None:
        if hover_col_list is not None:
            fig = px.line(
                df,
                x=x_col_name,
                y=y_col_name,
                color=color_col_name,
                hover_data=hover_col_list,
            ).for_each_trace(
                lambda t: t.update(name=t.name.replace(color_col_name + "=", ""))
            )
        else:
            fig = px.line(
                df, x=x_col_name, y=y_col_name, color=color_col_name
            ).for_each_trace(
                lambda t: t.update(name=t.name.replace(color_col_name + "=", ""))
            )
    else:
        if hover_col_list is not None:
            fig = px.line(df, x=x_col_name, y=color_col_name, hover_data=hover_col_list)
        else:
            fig = px.line(df, x=x_col_name, y=color_col_name)

    return fig



from uuid import uuid4
import pandas as pd


def fill_na_formater_and_checker(func):
    """
    Decorator for fill na functions
    """

    def inner(*args, **kwargs):
        fill_na_dict = func(*args, **kwargs)

        # Unpack arg
        df = fill_na_dict["df_original"]
        df_transformed = fill_na_dict["df_transformed"]
        groupby_cols = fill_na_dict["groupby_cols"]
        measure_col = fill_na_dict["measure_col"]

        # Structure output
        impute_instance = impute_class()
        df_output = impute_instance.merge_with_original_df_and_fill_na(
            _df_original=df,
            _df_transformed=df_transformed,
            groupby_cols=groupby_cols,
            measure_col=measure_col,
        )
        df_output = impute_instance.drop_original_measure_col_and_use_na_filled(
            _df=df_output, measure_col=measure_col
        )

        # Ensure output is of the right size
        if df_output.shape != df.shape:
            raise Exception("Output df is not of the right size")

        # Check if there are any more NA in measure_col
        if df_output[measure_col].isna().sum() > 0:
            print(
                f"There are still NA in {measure_col} - This will happen in cases where there are no non-na entry within a groupby object"
            )

        return df_output

    return inner


class impute_class:
    """
    1. groupby_rolling_mean_and_shift
    2. groupby_expanding_mean_and_shift
    3. groupby_fill
    4. use_summary_stat_of_groupby_to_fill_na
    5. fill_na_in_id_col
    """

    def __init__(self) -> None:
        pass

    def groupby_fill(
        self,
        df: pd.DataFrame,
        date_col: str,
        groupby_cols: list,
        measure_col: str,
        strategy: str = "bfill",
    ):
        df = df.copy()
        if strategy == "bfill":
            shift_pd = -1
        elif strategy == "ffill":
            shift_pd = 1
        else:
            raise ValueError("strategy must be either bfill or ffill")

        # Find the number of na values within each group
        mask = df[measure_col].isna()
        df_na_count_within_groupby = (
            df[mask].groupby(groupby_cols)[measure_col].size().reset_index()
        )
        df_na_count_within_groupby = df_na_count_within_groupby.rename(
            columns={measure_col: "na_count"}
        )
        max_na_count = int(df_na_count_within_groupby["na_count"].max())

        # rolling mean with 1 window
        for _ in range(max_na_count):
            df = self.groupby_rolling_mean_and_shift(
                _df=df,
                date_col=date_col,
                groupby_cols=groupby_cols,
                number_of_periods=1,
                measure_col=measure_col,
                min_pds=1,
                shift_pds=shift_pd,
            )
        return df

    def groupby_expanding_mean_and_shift(
        self,
        _df: pd.DataFrame,
        date_col: str,
        groupby_cols: list,
        measure_col: str,
        min_pds: int = 1,
    ):
        df = _df.copy()

        # Find number of periods - this is the window size
        number_of_periods = df[date_col].nunique() + 1

        return self.groupby_rolling_mean_and_shift(
            _df=df,
            date_col=date_col,
            groupby_cols=groupby_cols,
            number_of_periods=number_of_periods,
            measure_col=measure_col,
            min_pds=min_pds,
        )

    def check_if_groupby_results_in_single_value(self, _df_grouper) -> None:
        df_check = (
            _df_grouper.size()
            .reset_index()
            .rename(columns={0: "number_of_entries_per_groupby"})
        )
        max_count = df_check["number_of_entries_per_groupby"].max()
        if max_count > 1:
            raise Exception(
                "Warning: There are multiple entries per groupby. Check for duplicates or consider aggregating the data"
            )
        return None

    @fill_na_formater_and_checker
    def groupby_rolling_mean_and_shift(
        self,
        _df: pd.DataFrame,
        date_col: str,
        groupby_cols: list,
        number_of_periods: int,
        measure_col: str,
        min_pds: int = 1,
        shift_pds: int = 1,
    ):
        df = _df.copy()
        measure_col_transform_interim = measure_col + "_" + "rolling_mean"
        measure_col_transform_final = "transformed"

        # Make sure the data is in shape to perform the operation
        df_grouper_checker = df.groupby(groupby_cols + [date_col])
        self.check_if_groupby_results_in_single_value(_df_grouper=df_grouper_checker)

        # Rolling mean and shift
        df_grouper = df.set_index([date_col]).groupby(groupby_cols)
        df_rolling_mean = (
            df_grouper.rolling(window=number_of_periods, min_periods=min_pds)[
                measure_col
            ]
            .mean()
            .reset_index()
        )
        df_rolling_mean = df_rolling_mean.rename(
            columns={measure_col: measure_col_transform_interim}
        )
        df_rolling_mean[measure_col_transform_final] = df_rolling_mean.groupby(
            groupby_cols
        )[measure_col_transform_interim].shift(periods=shift_pds)

        # Structure dict and return dict
        fill_na_dict = {}
        fill_na_dict["df_original"] = df
        fill_na_dict["df_transformed"] = df_rolling_mean
        fill_na_dict["groupby_cols"] = groupby_cols + [date_col]
        fill_na_dict["measure_col"] = measure_col
        return fill_na_dict

    def merge_with_original_df_and_fill_na(
        self,
        _df_original: pd.DataFrame,
        _df_transformed: pd.DataFrame,
        groupby_cols: list,
        measure_col: str,
    ):
        df = _df_original.copy()
        df_transformed = _df_transformed.copy()
        measure_col_transform_final = "transformed"
        measure_col_na_filled = measure_col + "_na_filled"

        # Merge with original df
        df = df.merge(
            df_transformed[groupby_cols + [measure_col_transform_final]],
            on=groupby_cols,
            how="left",
        )
        df[measure_col_na_filled] = df[measure_col]

        # Fill na with rolling mean shifted value
        mask = df[measure_col].isna()
        df.loc[mask, measure_col_na_filled] = df.loc[mask, measure_col_transform_final]

        return df

    def drop_original_measure_col_and_use_na_filled(
        self, _df: pd.DataFrame, measure_col: str
    ):
        df = _df.copy()
        measure_col_na_filled = measure_col + "_na_filled"
        measure_col_transform_final = "transformed"

        df = df.drop(columns=[measure_col, measure_col_transform_final])
        df = df.rename(columns={measure_col_na_filled: measure_col})
        return df

    @fill_na_formater_and_checker
    def use_summary_stat_of_groupby_to_fill_na(
        self,
        _df: pd.DataFrame,
        groupby_cols: list,
        measure_col: str,
        summary_stat: str = "mean",
    ):
        df = _df.copy()

        # Perform groupby operation
        df_grouper = df.groupby(groupby_cols)

        # calculate summary stat
        # Note: Skip NA is True by default
        measure_col_transform_final = "transformed"
        df_summary = df_grouper[measure_col].agg(summary_stat).reset_index()
        df_summary = df_summary.rename(
            columns={measure_col: measure_col_transform_final}
        )

        # Structure dict and return dict
        fill_na_dict = {}
        fill_na_dict["df_original"] = df
        fill_na_dict["df_transformed"] = df_summary
        fill_na_dict["groupby_cols"] = groupby_cols
        fill_na_dict["measure_col"] = measure_col

        return fill_na_dict

    def fill_na_in_id_col(self, _df: pd.DataFrame, id_col: str):
        df = _df.copy()
        uuid_col = "uuid"

        # Generate a uid column
        df[uuid_col] = df.index.map(lambda _: uuid4().hex)

        # Fill na in id_col with values in uuid_col
        mask = df[id_col].isna()
        df.loc[mask, id_col] = df.loc[mask, uuid_col]

        # Make sure the filled values are unique and are not already a part of id_col
        uid_entered_values = list(df.loc[mask, id_col].values)
        if len(uid_entered_values) != len(set(uid_entered_values)):
            raise Exception(
                "Warning: There are duplicate values in the uid_entered_values"
            )
        unique_id_values = list(df.loc[~mask, id_col].unique().tolist())
        common_ids = list(set(unique_id_values).intersection(set(uid_entered_values)))
        if len(common_ids) > 0:
            raise Exception(
                f"Warning: There are common values between uid entered values and {id_col}"
            )

        # Drop uid column before returning df
        df = df.drop(columns=[uuid_col])

        return df



from eda_and_clean.eda import eda_class
from .fixtures.case_1.input import input_1_fixture
from .fixtures.case_1.output_eda import (
    output_1_dtypes_fixture,
    output_1_missing_values_fixture,
    output_1_downcasting_fixture,
    output_1_duplicates_fixture,
    output_1_data_analysis_fixture,
    output_1_summary_fixture,
    output_1_df_cdf,
)
from .utils import assert_dict, assert_dataframe

df = input_1_fixture()
eda_instance_1 = eda_class(_df=df)


def test_cdf():
    df_cdf_actual = eda_instance_1.calculate_empirical_cdf(_df=df, col="value")
    df_cdf_expected = output_1_df_cdf()
    assert_dataframe(actual=df_cdf_actual, expected=df_cdf_expected)


def test_summary():
    output, not_tested_keys = output_1_summary_fixture()
    assert_dict(
        actual=eda_instance_1.SUMMARY, expected=output, not_tested_keys=not_tested_keys
    )


def test_dtypes():
    output, not_tested_keys = output_1_dtypes_fixture()
    assert_dict(
        actual=eda_instance_1.DTYPES, expected=output, not_tested_keys=not_tested_keys
    )


def test_downcasting():
    output, not_tested_keys = output_1_downcasting_fixture()
    assert_dict(
        actual=eda_instance_1.DOWNCASTING,
        expected=output,
        not_tested_keys=not_tested_keys,
    )


def test_duplicates():
    output, not_tested_keys = output_1_duplicates_fixture()
    assert_dict(
        actual=eda_instance_1.DUPLICATES,
        expected=output,
        not_tested_keys=not_tested_keys,
    )


def test_missing_values():
    output, not_tested_keys = output_1_missing_values_fixture()
    assert_dict(
        actual=eda_instance_1.MISSING_VALUES,
        expected=output,
        not_tested_keys=not_tested_keys,
    )


def test_data_analysis():
    output, not_tested_keys = output_1_data_analysis_fixture()
    assert_dict(
        actual=eda_instance_1.DATA_ANALYSIS,
        expected=output,
        not_tested_keys=not_tested_keys,
    )



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="eda-and-clean",
    version="0.1.5",
    author="Aravind Ganesan",
    author_email="1988.aravind@gmail.com",
    description="A package of automation tools for EDA and cleaning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aravindganesan88/eda_and_clean",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "nltk==3.7",
        "numpy==1.21.6",
        "pandas==1.3.5",
        "plotly==5.11.0",
        "seaborn==0.12.1",
        "Nbformat==4.2.0",
    ],
    python_requires=">=3.7",
)
